/**
 * USOS SIGN Connector to biblioteka JS obsługująca komunikację między przeglądarką a usługą USOS
 * SIGN realizującą podpisywanie dokumentów oraz funkcjonalności związane z czytnikami kart
 * (certyfikaty i legitymacje).
 * 
 * Do komunikacji z usługą USOS SIGN wykorzystywany jest protokół WebSocket. Dane przesyłane między
 * przeglądarką są szyfrowane za pomocą kombinacji szyfrów RSA i AES. Do realizacji tego celu
 * wykorzystywana jest biblioteka Forge, którą należy załączyć do strony, na której ma być
 * wykorzystywany USOS SIGN Connector.
 * 
 * Autor: Krzysztof Więckowski
 */

UsosSignState = Object.freeze({
	NOT_CONNECTED : "NOT_CONNECTED",
	WAITING : "WAITING",
	WAITING_FOR_SERVER_KEY : "WAITING_FOR_SERVER_KEY",
	CARD_TEST : "CARD_TEST",
	XML_SIGNING : "XML_SIGNING",
	PDF_SIGNING : "PDF_SIGNING",
	KEY_NR : "KEY_NR",
	CONTACTLESS_NR : "CONTACTLESS_NR",
	CONTACT_NR : "CONTACT_NR",
	REQUEST_PIN : "REQUEST_PIN",
	PIN : "PIN",
	SHOW_CERT : "SHOW_CERT",
	SHOW_CERT_CARD : "SHOW_CERT_CARD",
	LOAD_CERT : "LOAD_CERT",
	LOAD_DATA : "LOAD_DATA",
	LOAD_DATA_WITH_LIBRARY_NR : "LOAD_DATA_WITH_LIBRARY_NR",
	PROLONG_CARD : "PROLONG_CARD",
	PROLONG_CARD_WITH_DATA : "PROLONG_CARD_WITH_DATA",
	PROLONG_CARD_WITH_STRUCTURE : "PROLONG_CARD_WITH_STRUCTURE",
	GENERATE_STRUCT : "GENERATE_STRUCT",
	LIBRARY_NR : "LIBRARY_NR",
	CHANGE_LIBRARY_NR : "CHANGE_LIBRARY_NR",
	CONFIG_VALUE : "CONFIG_VALUE",
	PERSON_CODE_CONTACTLESS_NR : "PERSON_CODE_CONTACTLESS_NR",
	PERSON_CODE_PRINT : "PERSON_CODE_PRINT",
	PERSON_CONTACT_NR : "PERSON_CONTACT_NR",
	PERSON_CONTACTLESS_NR : "PERSON_CONTACTLESS_NR",
	PERSON_INIT : "PERSON_INIT",
	PERSON_INIT_CONTACT : "PERSON_INIT_CONTACT",
	PRINT_PREP : "PRINT_PREP",
	PRINT_TEMPLATE : "PRINT_TEMPLATE",
	SUCCESS : "SUCCESS",
	BUSY : "BUSY"
});

function UsosSignConnector() {
	this.clientPrivateKey = null;
	this.servicePublicKey = null;
	this.state = UsosSignState.NOT_CONNECTED;
	this.cipherType = 'AES-CBC';
	this.onResponseFnc = {};
	this.waitForConnection.waitTime = 0;
	this.wrapper = window;
}

/**
 * Inicjalizacja połączenia z usługą USOS SIGN. W przypadku wystąpienia problemów z połączeniem,
 * powinno się wyświetlić stosowne ostrzeżenie.
 * 
 * @param privateKeyString
 *            prywatny klucz, który będzie używany do odszyfrowywania komunikatów otrzymywanych z
 *            usługi. Parametr wymagany.
 * @param host
 *            parametr określający adres i port, na którym należy ustanowić połączenie z usługą USOS
 *            SIGN. Parametr opcjonalny, domyślnie ustawiony na "localhost:8025".
 */
UsosSignConnector.prototype.connect = function(privateKeyString, host) {
	if (typeof forge == 'undefined') {
		alert("Brak biblioteki Forge.\nSprawdź poprawność załączonych plików oraz kolejność ich importowania.");
		return;
	}

	if (!privateKeyString) {
		alert("Pusty klucz prywatny. Sprawdź konfigurację serwera.");
		return;
	}

	if ("WebSocket" in window) {
		host = typeof host != 'undefined' ? host : "localhost:8025";
		if (host.slice(-1) != '/') {
			host += "/";
		}

		if (this.ws && this.ws.readyState === 1) {
			if (this.state == UsosSignState.WAITING) {
				this.disconnect();
				setTimeout(this.connect.bind(this, privateKeyString, host), 50);
				return;
			} else {
				var msg = "Połączenie w użyciu! Nie można przerwać połączenia.";
				alert(msg);
				throw msg;
			}
		}

		// otwieramy WebSocket
		this.ws = new WebSocket("wss://" + host + "card");

		var connector = this;

		/**
		 * Wyzwalacz uruchamiany po rozpoczęciu połączenia z usługą.
		 */
		this.ws.onopen = function() {
			connector.state = UsosSignState.WAITING_FOR_SERVER_KEY;
		};

		/**
		 * Wyzwalacz uruchamiany po otrzymaniu wiadomości od usługi.
		 */
		this.ws.onmessage = function(evt) {
			var receivedMsg = evt.data;
			var obj;
			if (connector.state == UsosSignState.WAITING_FOR_SERVER_KEY) {
				// wczytanie klucza publicznego serwera
				obj = JSON.parse(receivedMsg);
				connector.servicePublicKey = forge.pki.publicKeyFromPem(obj.publicKey);
				connector.state = UsosSignState.WAITING;
			} else if (connector.state == UsosSignState.BUSY) {
				var decryptedMessage = connector.decryptMessage(receivedMsg);
				obj = JSON.parse(decryptedMessage);
			} else {
				var decryptedMessage = connector.decryptMessage(receivedMsg);
				obj = JSON.parse(decryptedMessage);
				if (obj.type == "ERROR") {
					alert(obj.errorDesc);
				} else {
					alert(decryptedMessage);
				}
			}
			if (typeof (obj.id) != 'undefined'
					&& typeof (connector.onResponseFnc[obj.id]) == 'function') {
				var returnFunc = connector.onResponseFnc[obj.id];
				if (!(obj.type == "WARNING" || obj.type == "STATUS" || obj.type == "REQUEST_PIN" || (obj.type == "ERROR" && !obj.fatal))) {
					delete connector.onResponseFnc[obj.id];
					connector.state = UsosSignState.WAITING;
				}
				returnFunc.apply(connector.wrapper, [ obj ]);
			}
		};

		/**
		 * Wyzwalacz uruchamiany po zakończeniu połączenia z usługą.
		 */
		this.ws.onclose = function(event) {
			if (connector.state != UsosSignState.NOT_CONNECTED) {
				if (!event.wasClean) {
					alert("Połączenie z usługą kart i podpisu zostało niespodziewanie zakończone.");
				}
				connector.state = UsosSignState.NOT_CONNECTED;
			}
			connector.ws = null;
		};

		this.ws.onerror = function(event) {
			throw JSON.stringify(event);
		}

		// wczytujemy klucz prywatny
		try {
			this.clientPrivateKey = forge.pki.privateKeyFromPem(privateKeyString);
		} catch (e) {
			throw "Niepoprawny format klucza prywatnego aplikacji";
		}
	} else {
		var noWebSocket = "Oprogramowanie nie wspiera technologii WebSocket. Zaktualizuj przeglądarkę do najnowszej wersji.";
		alert(noWebSocket);
		throw noWebSocket;
	}

}

/**
 * Zakończenie połączenia z usługą - wyczyszczenie zmiennych i zamknięcie WebSocket.
 */
UsosSignConnector.prototype.disconnect = function() {
	if (this.ws && this.ws.readyState === 1) {
		if (this.state == UsosSignState.WAITING) {
			this.ws.close(1000);
			this.clientPrivateKey = null;
			this.servicePublicKey = null;
			this.onResponseFnc = {};
		} else {
			throw "Połączenie w użyciu! Nie można przerwać połączenia.";
		}
	}
}

/**
 * Sprawdza czy połączenie z usługą USOS SIGN zostało poprawnie ustanowione.
 */
UsosSignConnector.prototype.checkConnection = function() {
	if (!this.ws || this.ws.readyState != 1 || this.state == UsosSignState.NOT_CONNECTED
			|| this.state == UsosSignState.WAITING_FOR_SERVER_KEY) {
		return -1;
	}
	return 0;
}

/**
 * Funkcja służąca do oczekiwania na rozpoczęcie połączenia z usługą USOS SIGN.
 * 
 * Ponieważ inicjalizacja połączenia wykonywana jest asynchronicznie i może chwilę potrwać, nie
 * możemy od razu wysyłać wiadomości przez WebSocket. W związku z tym, poniższa funkcja sprawdza w
 * określonych odstępach czasu czy połączenie jest już dostępne. Jeśli na połączenie oczekujemy
 * dłużej niż ok. 5 sekund, zwracany jest błąd (przyjmujemy wtedy, że połączenia nie udało się
 * ustanowić).
 * 
 * @param callback
 *            funkcja wywoływana, kiedy połączenie zostanie ustanowione
 * @param interval
 *            czas w milisekundach, jaki ma mijać między kolejnymi sprawdzeniami, czy połączenie
 *            zostało ustanowione
 */
UsosSignConnector.prototype.waitForConnection = function(callback, interval) {
	this.waitForConnection.waitTime += interval;
	if (this.checkConnection() == 0) {
		callback();
		this.waitForConnection.waitTime = 0;
	} else {
		var that = this;
		if (this.waitForConnection.waitTime < 5000) {
			setTimeout(function() {
				that.waitForConnection(callback, interval);
			}, interval);
		} else {
			var msg = "Brak połączenia z usługą kart i podpisu. Sprawdź czy usługa na pewno jest włączona.";
			alert(msg);
			throw msg;
		}
	}
}

/**
 * Odszyfrowuje dane algorytmem RSAES PKCS#1 v1.5 przy użyciu klucza prywatnego aplikacji.
 * 
 * @param encodedData
 *            tablica bajtów zawierająca dane do odszyfrowania.
 */
UsosSignConnector.prototype.decryptKey = function(encodedData) {
	return this.clientPrivateKey.decrypt(encodedData);
}

/**
 * Szyfruje dane algorytmem RSAES PKCS#1 v1.5 przy użyciu klucza publicznego usługi.
 * 
 * @param encodedData
 *            tablica bajtów zawierająca dane do zaszyfrowania. Ze względu na ograniczenia algorytmu
 *            RSA, tablica musi mieć rozmiar rzędu maksymalnie kilkudziesięciu bajtów. Funkcja
 *            wykorzystywana jedynie do szyfrowania 16-bitowego klucza AES, nie powinna być
 *            wykorzystywana w innym celu.
 */
UsosSignConnector.prototype.encryptKey = function(encodedData) {
	return this.servicePublicKey.encrypt(encodedData);
}

/**
 * Odszyfrowuje wiadomość otrzymaną z serwera. W pierwszym kroku wiadomość jest parsowana jako JSON.
 * Otrzymujemy w ten sposób klucz, wektor inicjujący oraz zaszyfrowane dane. Najpierw odszyfrowujemy
 * klucz przy użyciu algorytmu RSA i klucza prywatnego aplikacji. W rezultacie otrzymujemy klucz
 * AES, który wraz z posiadanym wektorem inicjującym wykorzystujemy do odszyfrowania danych przy
 * użyciu algorytmu AES-CBC.
 * 
 * @param message
 *            wiadomość odebrana od serwera w postaci stringa
 */
UsosSignConnector.prototype.decryptMessage = function(message) {
	try {
		var obj = JSON.parse(message);
		var key = this.decryptKey(forge.util.decode64(obj.encodedKey));
		var initVector = forge.util.decode64(obj.initVector);
		var encryptedData = forge.util.decode64(obj.encodedData);
		var decipher = forge.cipher.createDecipher(this.cipherType, key);
		decipher.start({
			iv : initVector
		});
		decipher.update(forge.util.createBuffer(encryptedData));
		decipher.finish();
		var data = decipher.output.getBytes();
		return forge.util.decodeUtf8(data);
	} catch (e) {
		alert("Wystąpił problem z wiadomością wysłaną z usługi. Proszę spróbować ponownie.\n\n" + e)
		throw e;
	}
}

/**
 * Szyfruje wiadomość przed wysłaniem jej do serwera.
 * 
 * Najpierw generujemy klucz AES oraz wektor inicjujący. Przy ich użyciu, szyfrujemy dane szyfrem
 * AES-CBC. Wygenerowany klucz AES szyfrujemy za pomocą algorytmu RSA i klucza publicznego usługi.
 * 
 * Wiadomość dla serwera zawiera więc zaszyfrowany klucz, wektor inicjujący oraz zaszyfrowane dane.
 * 
 * @param message
 *            wiadomość odebrana od serwera w postaci stringa
 */
UsosSignConnector.prototype.encryptMessage = function(data) {
	var message = {};
	var key = forge.random.getBytesSync(16);
	var iv = forge.random.getBytesSync(16);
	var cipher = forge.cipher.createCipher(this.cipherType, key);
	cipher.start({
		iv : iv
	});
	cipher.update(forge.util.createBuffer(forge.util.encodeUtf8(data)));
	cipher.finish();
	var bytes = forge.util.hexToBytes(cipher.output.toHex());
	message.encodedData = forge.util.encode64(bytes);
	message.encodedKey = forge.util.encode64(this.encryptKey(key));
	message.initVector = forge.util.encode64(iv);
	return message;
}

/**
 * Wysyła wiadomość do usługi USOS SIGN, uprzednia odpowiednio ją szyfrując.
 */
UsosSignConnector.prototype.sendMessage = function(message, returnFunc) {
	if (this.checkConnection() != 0) {
		alert("Brak połączenia z usługą kart i podpisu");
		return;
	}
	if (typeof (returnFunc) == 'function') {
		message.id = forge.util.encode64(forge.random.getBytesSync(8));
		this.onResponseFnc[message.id] = returnFunc;
	}
	var newMessage = this.encryptMessage(JSON.stringify(message));
	this.ws.send(JSON.stringify(newMessage));
}

/**
 * Wysyła PIN karty do usługi USOS SIGN.
 */
UsosSignConnector.prototype.sendPIN = function(pin) {
	var msg = {};
	msg.PIN = pin;
	msg.type = UsosSignState.PIN;
	var connector = this;
	this.waitForConnection(function() {
		connector.sendMessage(msg, null);
	}, 50);
}

/**
 * Wyświetla podany certyfikat w okienku systemowym.
 * 
 * @param cert
 *            ciąg znaków zawierający certyfikat
 * @param returnFunc
 *            funkcja wywoływana przy otrzymaniu wiadomości z usługi
 */
UsosSignConnector.prototype.showCert = function(cert, returnFunc) {
	var msg = {};
	msg.content = cert;
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.SHOW_CERT);
}

/**
 * Wyświetla certyfikat z karty w okienku systemowym.
 * 
 * @param returnFunc
 *            funkcja wywoływana przy otrzymaniu wiadomości z usługi
 */
UsosSignConnector.prototype.showCertCard = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.SHOW_CERT_CARD);
}

/**
 * Wysyła XML-a do usługi USOS SIGN w celu podpisania go.
 */
UsosSignConnector.prototype.signXml = function(xml, returnFunc) {
	this.simpleOperationWithData(xml, returnFunc, UsosSignState.XML_SIGNING);
}

/**
 * Wysyła plik PDF do usługi USOS SIGN w celu podpisania go.
 */
UsosSignConnector.prototype.signPdf = function(pdf, returnFunc) {
	this.simpleOperationWithData(pdf, returnFunc, UsosSignState.PDF_SIGNING);
}

/**
 * Funkcja pomocnicza dla operacji, które poza typem nie przekazują dodatkowych parametrów do
 * usługi.
 * 
 * @param returnFunc
 *            funkcja wywoływana po otrzymaniu odpowiedzi z USOS SIGN
 * @param type
 *            typ operacji, która ma zostać wykonana
 */
UsosSignConnector.prototype.simpleOperation = function(returnFunc, type) {
	var msg = {};
	this.simpleOperationWithData(msg, returnFunc, type);
}

/**
 * Funkcja pomocnicza dla operacji, które przekazują obiekt z danymi do usługi.
 * 
 * @param msg
 *            obiekt przesyłany do usługi
 * @param returnFunc
 *            funkcja wywoływana po otrzymaniu odpowiedzi z USOS SIGN
 * @param type
 *            typ operacji, która ma zostać wykonana
 */
UsosSignConnector.prototype.simpleOperationWithData = function(msg, returnFunc, type) {
	var connector = this;
	this.waitForConnection(function() {
		connector.state = UsosSignState.BUSY;
		msg.type = type;
		connector.sendMessage(msg, returnFunc);
	}, 50);
}

/**
 * Testuje poprawność konfiguracji usługi USOS SIGN.
 */
UsosSignConnector.prototype.cardTest = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.CARD_TEST);
}

/**
 * Zwraca numer aktualnego klucza z pliku els_config.ini.
 */
UsosSignConnector.prototype.keyNr = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.KEY_NR);
}

/**
 * Zapisuje na legitymacji nowy numer karty bibliotecznej.
 */
UsosSignConnector.prototype.changeLibraryNr = function(msg, returnFunc) {
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.CHANGE_LIBRARY_NR);
}

/**
 * Zwraca numer bezstykowy karty włożonej do czytnika.
 */
UsosSignConnector.prototype.contactlessNr = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.CONTACTLESS_NR);
}

/**
 * Zwraca numer stykowy karty włożonej do czytnika.
 */
UsosSignConnector.prototype.contactNr = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.CONTACT_NR);
}

/**
 * Zwraca numer biblioteczny karty włożonej do czytnika.
 */
UsosSignConnector.prototype.libraryNr = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.LIBRARY_NR);
}

/**
 * Generuje podpisaną strukturę przy pomocy przekazanych danych oraz włożonego do czytnika
 * certyfikatu
 */
UsosSignConnector.prototype.generateStruct = function(msg, returnFunc) {
	this.prolongCardGeneric(msg, returnFunc, UsosSignState.GENERATE_STRUCT);
}

/**
 * Przedłuża ważność legitymacji włożonej do czytnika
 */
UsosSignConnector.prototype.prolongCard = function(msg, returnFunc) {
	this.prolongCardGeneric(msg, returnFunc, UsosSignState.PROLONG_CARD);
}

/**
 * Przedłuża ważność legitymacji włożonej do czytnika przy użyciu przekazanych danych
 */
UsosSignConnector.prototype.prolongCardWithData = function(msg, returnFunc) {
	this.prolongCardGeneric(msg, returnFunc, UsosSignState.PROLONG_CARD_WITH_DATA);
}

/**
 * Przedłuża ważność legitymacji włożonej do czytnika za pomocą gotowej struktury
 */
UsosSignConnector.prototype.prolongCardWithStructure = function(msg, returnFunc) {
	this.prolongCardGeneric(msg, returnFunc, UsosSignState.PROLONG_CARD_WITH_STRUCTURE);
}

/**
 * Przedłuża ważność legitymacji włożonej do czytnika w wybrany sposób
 * 
 * @param msg
 *            wiadomość zawierająca wszystkie dane potrzebne do przedłużenia legitymacji w wybranym
 *            trybie
 * @param returnFunc
 *            funkcja wywoływana przy otrzymaniu wiadomości z usługi
 * @param type
 *            tryb w jakim legitymacja ma zostać przedłużona: 1. PROLONG_CARD - standardowe
 *            przedłużenie, 2. PROLONG_CARD_WITH_DATA - przedłużenie z podaniem wszystkich danych
 *            (szybsze - nie są wczytywane dane) 3. PROLONG_CARD_WITH_STRUCTURE - przedłużenie przy
 *            użyciu gotowej struktury
 */
UsosSignConnector.prototype.prolongCardGeneric = function(msg, returnFunc, type) {
	this.simpleOperationWithData(msg, returnFunc, type);
}

/**
 * Wczytuje certyfikat z karty.
 * 
 * @param returnFunc
 *            funkcja wywoływana przy otrzymaniu wiadomości z usługi
 */
UsosSignConnector.prototype.loadCert = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.LOAD_CERT);
}

/**
 * Zwraca dane osoby zapisane na legitymacji.
 */
UsosSignConnector.prototype.loadCardData = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.LOAD_DATA);
}

/**
 * Zwraca dane osoby zapisane na legitymacji wraz z numerem karty bibliotecznej.
 */
UsosSignConnector.prototype.loadCardDataWithLibNr = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.LOAD_DATA_WITH_LIBRARY_NR);
}

/**
 * Zwraca wartość parametru konfiguracyjnego z pliku els_config.ini.
 */
UsosSignConnector.prototype.configValue = function(msg, returnFunc) {
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.CONFIG_VALUE);
}

/**
 * Personalizacja układu stykowego i drukowanie legitymacji.
 */
UsosSignConnector.prototype.personCodePrint = function(msg, returnFunc) {
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.PERSON_CODE_PRINT);
}

/**
 * Personalizacja układu bezstykowego przed drukowaniem.
 */
UsosSignConnector.prototype.personCodeContactlessNr = function(msg, returnFunc) {
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.PERSON_CODE_CONTACTLESS_NR);
}

/**
 * Zwraca numer stykowy karty i dane układu stykowego przed drukowaniem.
 */
UsosSignConnector.prototype.personContactNr = function(repeat, returnFunc) {
	msg = {};
	msg.repeat = repeat;
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.PERSON_CONTACT_NR);
}

/**
 * Zwraca numer bezstykowy karty przed drukowaniem.
 */
UsosSignConnector.prototype.personContactlessNr = function(repeat, returnFunc) {
	msg = {};
	msg.repeat = repeat;
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.PERSON_CONTACTLESS_NR);
}

/**
 * Przygotowuje drukarkę do personalizacji karty.
 */
UsosSignConnector.prototype.personInit= function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.PERSON_INIT);
}

/**
 * Inicjuje układ stykowy na karcie.
 */
UsosSignConnector.prototype.personInitContact= function(msg, returnFunc) {
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.PERSON_INIT_CONTACT);
}

/**
 * Odczytanie numeru bezstykowego z drukarki.
 */
UsosSignConnector.prototype.printPrep = function(returnFunc) {
	this.simpleOperation(returnFunc, UsosSignState.PRINT_PREP);
}

/**
 * Inicjacja układu stykowego i drukowanie legitymacji według szablonu.
 */
UsosSignConnector.prototype.printTemplate = function(msg, returnFunc) {
	this.simpleOperationWithData(msg, returnFunc, UsosSignState.PRINT_TEMPLATE);
}

var UsosSign;
if (UsosSign == null) {
	UsosSign = new UsosSignConnector();
}